/*++++++++++++++
.IDENTIFICATION ipname.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    Unix
.KEYWORDS       
.VERSION  1.0   25-Apr-1995
.VERSION  1.1   09-Oct-2005: Accept list in input
.COMMENTS       From example on Solaris
---------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#define is_comment(c)	((c)=='#') || ((c)=='"') || ((c)=='\\')

static char usage[] = "\
Usage: ipname [-s] [-num | IP]" ;

static char help[] = "\
    -s: silently ignore non-found IP numbers\n\
  -num: number indicated the word position of IP number (e.g. -7 for xferlog)\n\
   IP : Numeric IP number\n\
" ;

typedef struct aNode { 
   struct aNode *next;
   u_long ip ;
   char name[4];	/* Actual size differs */
} NODE ;

static NODE *list[256] ;

static char verbop = 0;
static char opt_s = 0;

static NODE *find_ip (char *host)
/*++++++++++++++++
.PURPOSE  Find the name corresponding to an IP number
.RETURNS  The corresponding name
.REMARKS  Added if not yet existing
-----------------*/
{
  NODE *n, *p ;
  char *h, *hnum, *hnam ;
  struct hostent *hp;
  /* struct in_addr inaddr ; */
  u_long addr ;	
  int i, x ;

    hnum = hnam = (char *)0 ;
    for (h=host; isgraph(*h); h++) ;

    if (isdigit(*host)) {	/* Must look for the host name ... */
	hnum = host ;
    }
    else {
	hnam = host;
        while (isgraph(*h)) h++ ; while (isspace(*h)) h++ ;
	if (*h == '(') h++ ;
	hnum = h ;
	while (isdigit(*h) || (*h == '.')) h++;
    }
    if (!hnum) return((NODE *)0) ;

    /* Convert the IP number, h = end of hnum string */
    x = *h; *h = 0; 
    addr = inet_addr(hnum); 
    *h = x ;
    if ((int)(addr) == -1) {
	if (opt_s) ;
        else fprintf(stderr, "****Bad IP: %s\n", hnum);
	return ((NODE *)0) ;
    }

    /* Look: is the IP number already saved ? */
    i = ((addr>>24) + (addr>>8) + (addr&0xff)) & 0xff ;
    n = p = list[i] ;
    while (n) {
        p = n ;
	if (n->ip == addr) break ;
	n = n -> next ;
    }
    if (n) 		/* It's there !! */
        return(n) ;

    /* Not yet Found... Is the host name available ? */
    if (!hnam) {
        hp = gethostbyaddr((char *)&addr, sizeof (addr), AF_INET);
        if (hp) hnam = hp->h_name ;
	else {
	    x = *h; *h = 0;
	    if (opt_s) ; else fprintf(stderr, 
	      "++++host information for %08lX (%s) not found\n", 
	      addr, hnum);
	    *h = x ;
	    hnam = hnum ;
	}
    }

    /* Allocate the new corespondance */
    for (h=hnam; isgraph(*h); h++) ;
    x = h-hnam ;
    n = (NODE *)malloc(sizeof(NODE) + x - 3) ;
    n->ip = addr ;
    memcpy(n->name, hnam, x) ;
    n->name[x] = 0 ;
    n->next = (NODE *)0 ;
    if (p) p->next = n ;
    else list[i] = n ;

    return(n) ;
}


int main(int argc, char **argv) 
{
  int found = 0;
  int word =  1;
  char *arg, *p;
  char line[4096];
  int i, len, cont;
  NODE *n;

    while (--argc>0) {
	arg = *++argv;
	if (*arg == '-') switch(arg[1]) {
	  case 's':
	    opt_s = 1;
	    continue;
	  case 'v':
	    verbop = 1;
	    continue;
	  case 'h':	/* Help */
	    printf("%s", help);
	    exit(0);
	  default:
	    if (isdigit(arg[1])) {
		word = atoi(arg+1);	/* Offset in word */
		continue;
	    }
	    fprintf(stderr, "****%s\n", usage);
	    exit(1);
	}

	/* A single argument: we assume it's an IP number */
	found++;
	if ((n = find_ip(arg))) printf("%s\n", n->name);
    }

    if (found>0) exit(0);

    /* Take names from stdin */

    cont = 0;
    while (fgets(line, sizeof(line), stdin)) {
	len = strlen(line) -1;
	if (cont) {		/* A continuation */
	    printf("%s", line);
	    if (line[len] == '\n') cont = 0;
	    continue;
	}
	found++;

	cont = line[len] != '\n' ;
	if (is_comment(line[0])) continue;

	/* Locate the word */
	for (p=line; isspace(*p); p++) ;
	for (i=word; --i>0; ) {
	    while (isgraph(*p)) p++;
	    while (isspace(*p)) p++;
	}
	if ((n = find_ip(p))) {
	    fwrite(line, 1, p-line, stdout);
	    printf("%s", n->name);
	    while (isgraph(*p)) p++;
	    printf("%s", p);
	}
	else {
	    printf("%s", line);
	}
    }
    exit (0);
}
