/*++++++++++++++
.IDENTIFICATION wwwabs.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    
.KEYWORDS       HTML
.VERSION  1.0   18-Oct-1998
.VERSION  1.1   13-Jun-2007: strndup & gnu
.COMMENTS       Replace relative anchors by absolute ones in a HTML source.
---------------*/

#define VERSION "1.0"

#ifdef __GNUC__		/* Added V1.1 */
#define _GNU_SOURCE
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

static char verbop ;		/* Verbose Option		*/
static FILE *infile ;

#define GetChar()	fgetc(infile)
#define UnGetChar(c)	ungetc(c, infile)
#define isquote(c)	((c) == '"') || ((c) == '\'') || ((c) == '`')

static char *addr[] = {
    "HREF=", "SRC=", (char *)0
} ;

static char *url[5] ;		/* Components of an URL	*/
#define URL_PROTOCOL	0
#define URL_NODE	1
#define URL_DOMAIN	2
#define URL_PORT	3
#define URL_PATH	4

static char usage[] = 
	"Usage: wwwabs [-v] URL [file...]\n";

static char help[] = "\
This program replaces the relative anchors (text or image) into absolute ones\n\
typically to save an HTML result as a stand-alone file.\n\
  URL: the resource locator used to generate the text, e.g. http://...\n\
" ;

/*==================================================================
		BSD Utilities
 *==================================================================*/
#ifdef __MSDOS__ 	/* For tests */
static int gethostname(buf, len)	char *buf; int len ; {
    strcpy(buf, "PC-DOS.PC-DOMAIN");
    return(0);
}
char *getdomain() {
    return(".PC-DOMAIN");
}
#else			/* == Standard Unix */
#include <sys/utsname.h>	/* for gethostname	*/
char *getdomain()
/*++++++++++++++++
.PURPOSE  Get name of domain (e.g. u-strasbg.fr)
.RETURNS  The domain (no starting dot)
.REMARKS  From /etc/resolv.conf
-----------------*/
{
  static char *domain = (char *)0 ;
  char buf[80], *p ;
  FILE *resolv ;

    if (domain) return(domain) ;

    if (!(resolv = fopen("/etc/resolv.conf", "r")))
	return("") ;
    while (fgets(buf, sizeof(buf), resolv)) {
	if (strncmp(buf, "domain", 6) != 0) continue ;
	p = buf + 6 ; while (isspace(*p)) p++ ;
	domain = strdup(p) ;
	for (p=domain; isgraph(*p); p++)  ;
	*p = 0 ;
	break ;
    }
    fclose(resolv) ;
    return(domain) ;
}

static int gethostname(host, len)
/*++++++++++++++++
.PURPOSE  Get name of host
.RETURNS  0 (OK) / 1 (truncated) /-1(error)
.REMARKS  BSD compatibility routine. Add the domain name if required.
-----------------*/
  char *host;	/* OUT: Name of host 	*/
  int len;	/* IN: Size of name	*/
{
  struct utsname name; 
  int status, i;
  char *domain, *p ;
    status = uname(&name);
    if (status < 0) return(status);

    for (p=name.nodename; *p && (*p != '.'); p++) ;
    if (*p) domain = "" ;
    else domain = getdomain() ;

    /* Copy the result */
    status = 0 ;
    i = strlen(name.nodename) ; 
    strncpy(host, name.nodename, len); 
    if (i >= len) host[i=len-1]=0, status = 1;

    /* Append the domain */
    if (domain && *domain) {
	p = host + i, len -= i ;
	if (len > 0) *(p++) = '.', len-- ;
	strncpy(p, domain, len) ;
	i = strlen(p) ;
	if (i >= len) p[i=len-1]=0, status = 1;
    }
    return(status);
}
#endif

/*==================================================================
		Internal Routines
 *==================================================================*/

static int strloc(text, c)
/*++++++++++++++++
.PURPOSE  Locate specified character
.RETURNS  Index of located char
-----------------*/
  char *text;     /* IN: String to interpret */
  int  c;         /* IN: Char to look for */
{
  char *s;
    for ( s = text; *s; s++)      if (*s == c)    break;
    return(s-text);
}

#ifndef __GNUC__	/* V1.1 */
static char *strndup(text, n)
/*++++++++++++++++
.PURPOSE  Keep a copy of a string
.RETURNS  Newly allocated string
-----------------*/
  char *text;     /* IN: String to interpret */
  int  n;         /* IN: Length to duplicate */
{
  char *s;
    if (n < 0) n = strlen(text) ;
    s = malloc(n+1) ;
    strncpy(s, text, n) ;
    s[n] = 0 ;
    return(s) ;
}
#endif			/* V1.1 */

static int cut_url(text, len5)
/*++++++++++++++++
.PURPOSE  Decompose the URL into parts
.RETURNS  Number of components (normally 4) / 0 when not applicable
		(nor http neither ftp protocol, or internal anchor)
-----------------*/
  char *text;	/* IN: String to interpret */
  int  *len5;	/* OUT: 5 numbers with lengths of components */
{
  int len, i ;
  char *p, *e ;
  
    if (*text == '#')		/* Internal Anchor */
	return(0) ;

    len = strloc(text, '?') ;
    p = text ; e = p + len ;
    i = 0 ;
    
    /* Protocol: only alphabetic */
    while (isalpha(*p)) p++ ;
    if (*p == ':') { 
	len5[i] = (++p-text) ; 
	if ((tolower(text[0]) != 'h') && (tolower(text[0]) != 'f'))
	    return(0) ;		/* Can't be http, nor ftp ...	*/
	text = p ; 
    }
    else len5[i] = 0 ;		/* Protocol */
    
    len5[++i] = 0 ;		/* Nodename */
    /* Hostname: starts by // */
    if ((p[0] == '/') && (p[1] == '/')) { 
    	p += 2 ;
    	while (isgraph(*p) && (p < e) && (*p != '.') 
    	    && (*p != '/') && (*p != ':')) p++ ;
    	len5[i] = (p-text) ; text = p ;
    	while (isgraph(*p) && (p < e) && (*p != '/') && (*p != ':')) p++ ;
    	len5[++i] = (p-text) ; text = p ;
    	if (*p == ':') {
    	    for (++p; isdigit(*p); p++) ;
    	    len5[++i] = (p-text) ; text = p ;
    	}
    	else len5[++i] = 0 ;
    }
    else {
        len5[++i] = 0 ;	/* Domain.. */
        len5[++i] = 0 ;	/* Port.... */
    }
    
    /* What remains is the path... */
    len5[++i] = (e-text) ;
    return(++i) ;
}
 
static int match_string(sep, buf, size)
/*++++++++++++++++
.PURPOSE  Match a complete string
.RETURNS  Last byte read (EOF or " or \n)
.REMARKS  Quote may be doubled to indicate a quote.
-----------------*/
  int  sep;    	/* IN: Char just read (' or ") */
  char *buf;	/* OUT: string stored 	*/
  int  size;	/* IN: Size of buf	*/
{
  int c=0;
  char *b, *e ;
    for (b=buf, e=buf+size-1; b < e; b++) {
        c = GetChar() ;
        if (c == sep) {		/* Another Quote */
            c = GetChar() ;
            if (c == sep) { *b = sep ; continue ; }
            UnGetChar(c) ;
            c = sep ;
            break ;
        }
        if (c == '\n') break ;
        if (c == EOF)  break ;
        *b = c ;
    }
    *b = 0 ;
    return(c) ;
}

static int match_tag()
/*++++++++++++++++
.PURPOSE  Match a complete tag, and replace the URL.
.RETURNS  Character found ('>' or EOF)
.REMARKS  '<' just read.
-----------------*/
{
  int c, i, n, change, len5[5];
  char name[12];
  char string[1025], *s ;
  char **a ;
  
    c = GetChar() ;
    if ((c == '!') || (c == '%') || (c == '&') || (c == '/'))	
	change = -1 ; 	/* Can't have an URL */
    else change = 0 ;


    while ((c != '>') && (c != EOF)) {
        while (isspace(c)) { putchar(c); c = GetChar() ; continue ; }
        n = change ;		/* Number of chars loaded for HREF= */
        while (isgraph(c)) {
            if (c == '>') break ;
            if (isquote(c)) {
		putchar(c) ;
                c = match_string(c, string, sizeof(string)) ;
		printf("%s", string) ;
                if (c != EOF) putchar(c), c = GetChar() ;
		n = -1 ;
                continue ;
            }
	    putchar(c) ;
	    if ((n >= 0) && (n < sizeof(name)-1)) name[n++] = toupper(c) ;
	    else { c = GetChar() ; n = -1; continue ; }
	    if (c != '=') { c = GetChar() ; continue ; }

	    /* A word terminated by = : is a HTML parameter */
	    name[n] = 0 ;
	    for (a = addr; *a && strcmp(*a, name); a++) ;
	    if (! *a) { c = GetChar() ; n = -1; continue ; }

	    /*** Here start replacement of Anchor ***/
	    c = GetChar()  ;
	    if (isquote(c)) {
		putchar(c) ;
		c = match_string(c, string, sizeof(string)) ;
	    }
	    else {	/* Get Word */
		for (i=0; (i<sizeof(string)-1) && isgraph(c); i++) {
		    if (c == '>') break ;
		    string[i] = c ;
		    c = GetChar() ;
		}
		string[i] = 0 ;
	    }
	    /*** Translate string into a complete Anchor */
	    if (cut_url(string, len5)) {
	        for (i=0, s=string; i<4; s += len5[i++]) {
		    if (len5[i] == 0) printf("%s", url[i]) ;
		    else fwrite(s, 1, len5[i], stdout) ;
	        }
	        if (*s != '/') printf("%s", url[4]) ;	/* PATH */
	        printf("%s", s) ;
	    }
	    else printf("%s", string) ;

	    if (isquote(c)) { putchar(c) ; c = GetChar(); }
	    n = -1 ;
        }
    }
    return(c) ;
}

/*==================================================================
		Execute on opened file
 *==================================================================*/
static void exec_abs()
/*++++++++++++++++
.PURPOSE  Execute translation relative --> absolute anchors on open file
.RETURNS  ---
-----------------*/
{
  int c;
    c = 0 ;
    while (c != EOF) {
	if (c) putchar(c) ;
	if (c == '<') c = match_tag();
	else c = GetChar() ;
    }
}

/*==================================================================
		Program
 *==================================================================*/
int main (argc, argv)
int argc; char **argv;
{
  int   len5[5] ;
  char  *p;
  char hostname[120] ;

	/* Check the Arguments */
  
    hostname[0] = hostname[1] = '/' ;
    gethostname(hostname+2, sizeof(hostname)-2);

    while (-- argc > 0) {
	p = *++argv;
	if (*p == '-') switch(*++p) {
	  case 'v': 
	    if (strncmp(p, "ver", 3) == 0) {
		printf("wwwabs (CDS), Version %s\n", VERSION);
		exit(0);
	    }
	    verbop = 1 ;
	    continue ;
	  case 'h':				/* Help */
	    printf("%s\n%s", usage, help) ;
	    exit(0) ;
	  default:
	    fprintf(stderr, "****Unknown option: %s\n", p-1) ;
	    exit(1) ;
	}
	else if (!url[0]) {			/* Get URL */
	    if (cut_url(p, len5) == 0) {
		fprintf(stderr, "****URL is neither HTTP nor FTP...\n%s\n",
		    usage) ;
		exit(1) ;
	    }
	    if (!len5[0]) url[0] = "http:" ;
	    else url[0] = strndup(p, len5[0]), p += len5[0] ;
	    if (!len5[1]) url[1] = strndup(hostname, strloc(hostname, '.')) ;
	    else url[1] = strndup(p, len5[1]), p += len5[1] ;
	    if (!len5[2]) url[2] = strndup(hostname+strloc(hostname, '.'), -1) ;
	    else url[2] = strndup(p, len5[2]), p += len5[2] ;
	    if (!len5[3]) url[3] = ""; /* ":80" */
	    else url[3] = strndup(p, len5[3]), p += len5[3] ;
	    if (!len5[4]) url[4] = "/" ;
	    else {	/* Complete the Path -- Remove the File */
		url[4] = strndup(p, len5[4]+1);
		for (p = url[4]+strlen(url[4])-1; (p>url[4]) && (*p != '/') ;
		     p--) ;
		*p = '/' ; p[1] = 0 ;
	    }
	    if (*url[3]) {	/* Attach port to Domain */
		p = url[2] ;
		url[2] = malloc(strlen(url[2])+strlen(url[3])+1) ;
		strcpy(url[2], p); strcat(url[2], url[3]) ;
		*url[3] = 0 ;
		free(p);
	    }
	    continue ;
	}

	/* A file. Open it */
	infile = fopen(p, "r") ;
	if (!infile) {
	    fprintf(stderr, "****"); perror(p); 
	    exit(1); 
	}
	/* Write a comment */
	printf("<!-- Input from file: %s -->\n", p) ;
	exec_abs() ;
	fclose(infile) ;
    }

    /* Execute on stdin if necessary */
    if (!infile) {
	if (!url[0]) {
	    fprintf(stderr, "****Missing URL\n%s", usage) ;
	    exit(1) ;
	}
	infile = stdin ;
        exec_abs() ;
    }

    exit(0);
}
